
/* 
   GAI Album v0.6 written by Jonas Aaberg <cja@gmx.net>
   
   Fancy changes between covers and %F code by greg@drakos7.net.
   AlbumArt support by Thomas Zajic <zlatko@gmx.at>

   Displays the current album that XMMS is playing.

   Released under GPL.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>  /* usleep */
#include <gai/gai.h>
#include "xmmsctrl.h"
#include "config.h"

/* from gai-bgswitcher by greg@drakos7.net */
#define DEFAULT_ANIMATE FALSE		/* smooth fade from one thumbnail to next */
#define DEFAULT_ANIM_TYPE 1			/* fade, slide1, slide 2, zoom, random */
#define ANIMATION_TYPES 5			/* fade, slide1, slide 2, zoom, random */
#define ANIMATION_METHOD GDK_INTERP_NEAREST  /*GDK_INTERP_NEAREST, GDK_INTERP_HYPER*/
#define ANIMATION_SPEED 1			/* mid-range speed for transitions */




/* Size +1 for all except tag, since we want things to end on zero */
typedef struct 
{
    unsigned char tag[3];
    unsigned char track[31];
    unsigned char artist[31];
    unsigned char album[31];
    unsigned char year[5];
    unsigned char comment[31];
    unsigned char genre[3];    
} ID3;


static GdkPixbuf *no_xmms, *naif;


static char *picname_format_front;
static char *picname_format_back;
static char *album_dir, *local_dir, *view_prg;

static int no_image, force_update = 0, front = 1;
static int pref_anim_type, pref_fancy_animation;	/*--*/

/* from gai-bgswitcher by greg@drakos7.net */

static char *radio_names[] = {"Fade", "Zoom", "Slide Right", "Slide Down", "Random", NULL};

static GaiPI album_pref[] = {{GAI_TEXTENTRY, "Format for front cover:", &picname_format_front, &picname_format_front},
			     {GAI_TEXTENTRY, "Format for back cover:", &picname_format_back, &picname_format_back},
			     {GAI_TEXT,      "%a means album name\n%p means performers name\n%y means year\n"
			      "%c means comment\n%F means the directory of the current played mp3"},
			     {GAI_TEXTENTRY, "Directory where the covers are stored:",&album_dir, &album_dir},
			     {GAI_TEXTENTRY, "Program used for viewing fullsize covers:", &view_prg, &view_prg},
			     {GAI_FRAME, "Transition Options:"},
			     {GAI_CHECKBUTTON, "Fancy Thumbnail Transition", &pref_fancy_animation, &pref_fancy_animation},
			     {GAI_RADIOBUTTON, radio_names, &pref_anim_type, &pref_anim_type},
			     {GAI_TEXT,       "Notice that in order to get the correct cover shown when playing CDs\n"
			      "and using cddb, the display format must begin with %p - %a - .."},
			     {GAI_FRAME_E},
			     {GAI_END}};


static char *help_text = "In the given directory, you must store front\n"
                         "covers of the album that XMMS is playing.\n"
                         "The mp3 files must have a valid ID3 v1.x tag.\n"
                         "?`*+\\/><|: are illegal signs in file names, so they're\n"
                         "translated to _ when looking for cover files.\n"
                         "Notice, that all is case sensitive and that a thumb image\n"
                         "is saved in ~/.gai-album, so if you want to update an image,\n"
                         "you have to remove it there too.";

void end_with_zero(char *str)
{
    int i;
    for(i=31;i>0;i--){
	if(str[i]==' ' || str[i]=='\0') 
	    str[i]='\0';
	else
	    break;
    }
}

void remove_ill(char *str)
{
    int i;
    for(i=0;i<30;i++){
	if(str[i]=='\0') break;
	if(str[i]=='?' ||
	   str[i]=='`' ||
	   str[i]=='*' ||
	   str[i]=='+' ||
	   str[i]=='\\' ||
	   str[i]=='/' ||
	   str[i]=='>' ||
	   str[i]=='<' ||
	   str[i]=='|' ||
	   str[i]==':') str[i]='_';
    }
}


int get_id3_tag(char *fname, ID3 *id3tag, int remove_illegal)
{
    FILE *f;
    int fsize;

    f = fopen(fname,"r");
    if(f<=0)
	return 1;
  
    fseek(f,0,SEEK_END);
    fsize=ftell(f);
    fsize-=128;

    fseek(f,fsize,SEEK_SET);
    fread(id3tag->tag,1,3,f);
    fread(id3tag->track,1,30,f);
    fread(id3tag->artist,1,30,f);
    fread(id3tag->album,1,30,f);
    fread(id3tag->year,1,4,f);
    fread(id3tag->comment,1,30,f);
    fread(id3tag->genre,1,2,f);

    fclose(f);

    if(id3tag->tag[0]=='T' && id3tag->tag[1]=='A' && id3tag->tag[2]=='G'){
	end_with_zero(id3tag->track);
	end_with_zero(id3tag->album);
	end_with_zero(id3tag->artist);
	end_with_zero(id3tag->comment);

	if(remove_illegal){
	    remove_ill(id3tag->track);
	    remove_ill(id3tag->album);
	    remove_ill(id3tag->artist);
	    remove_ill(id3tag->comment);
	}
    }
    else
	return 1;

    return 0;
}



char *album_get_pic_file(char *format, int remove_illegal)
{
    char *mp3_file=NULL, *pic_file=NULL, *lastslash=NULL, *dir=NULL;
    ID3 *id3tag;
    int i, j, need_tag=FALSE, has_tag=FALSE;


    if(xmms_remote_get_playlist_length(0)==0)
	return NULL;

    mp3_file = xmms_remote_get_playlist_file(0, xmms_remote_get_playlist_pos(0));

    if(mp3_file == NULL)
	return NULL;

    id3tag = g_malloc0(sizeof(ID3));

    /* Must have a valid mp3 extension only if path to image is tag-based */
    if(!strncasecmp(mp3_file+strlen(mp3_file)-4,".mp3",4)){
       if(!get_id3_tag(mp3_file, id3tag, remove_illegal))
	   has_tag=TRUE;
       else
	   g_free(id3tag);

    }
    else if(!strncasecmp(mp3_file+strlen(mp3_file)-4,".cda",4)){
	has_tag = TRUE;
	g_free(mp3_file);
	mp3_file = xmms_remote_get_playlist_title(0, xmms_remote_get_playlist_pos(0));
	/* CD format %p - %a - %n - %t */

	for(i=0;i<strlen(mp3_file);i++){
	    if(mp3_file[i] == ' ' && mp3_file[i+1] == '-' && mp3_file[i+2] == ' '){
		id3tag->artist[i] = '\0';
		break;
	    }
	    id3tag->artist[i] = mp3_file[i];
	}
	for(j=i+3;j<strlen(mp3_file);j++){
	    if(mp3_file[j] == ' ' && mp3_file[j+1] == '-' && mp3_file[j+2] == ' '){
		id3tag->album[j-i-3] = '\0';
		break;
	    }
	    id3tag->album[j-i-3] = mp3_file[j];
	}
    }


    pic_file = g_malloc0(2048);

    j = 0;
    for(i=0;i<strlen(format);i++){

	/* Album name */
	if(format[i]=='%' && format[i+1] == 'a'){
	    if(has_tag){
		strcpy(pic_file+j,id3tag->album);
		j+=strlen(id3tag->album);
	    }
	    i++;
	    need_tag=TRUE;
	    continue;
	}

	/* Performer name */
	if(format[i]=='%' && format[i+1] == 'p'){
	    if(has_tag){
		strcpy(pic_file+j,id3tag->artist);
		j+=strlen(id3tag->artist);
	    }
	    i++;
	    need_tag=TRUE;
	    continue;
	}


	/* Year */
	if(format[i]=='%' && format[i+1] == 'y'){
	    if(has_tag){
		strcpy(pic_file+j,id3tag->year);
		j+=strlen(id3tag->year);
	    }
	    i++;
	    need_tag=TRUE;
	    continue;
	}

	/* Track */
	if(format[i]=='%' && format[i+1] == 't'){
	    if(has_tag){
		strcpy(pic_file+j,id3tag->track);
		j+=strlen(id3tag->track);
	    }
	    i++;
	    need_tag=TRUE;
	    continue;
	}


	/* Comment */
	if(format[i]=='%' && format[i+1] == 'c'){
	    if(has_tag){
		strcpy(pic_file+j,id3tag->comment);
		j+=strlen(id3tag->comment);
	    }
	    i++;
	    need_tag=TRUE;
	    continue;
	}

	/* Directory of current mp3 */
	if(format[i]=='%' && format[i+1] == 'F'){
	    lastslash = g_strrstr(mp3_file,"/");
	    /* Last slash is a pointer inside mp3_file, so no g_free on it */
	    if(lastslash == NULL){
		g_free(mp3_file);
		g_free(id3tag);
		return NULL;
	    }
	    dir = g_strndup(mp3_file, strlen(mp3_file) - strlen(lastslash));


	    strcpy(pic_file+j,dir);
	    j+=strlen(dir);
	    i++;
	    g_free(dir);
	    continue;
	}

	pic_file[j] = format[i];
	j++;
    }
    
    pic_file[j]='\0';

    g_free(mp3_file);

    if(has_tag)
	g_free(id3tag);
    else {
        /* Last resort: if not looking for the image file location, try to ask  */
	/* xmms directly. This *should* not break anything, and works fine for  */
	/* detecting song changes for MP3s without tags, provided that the xmms */
	/* title format (Options -> Peferences -> Title -> Title format) is set */
	/* to some reasonable value - which it probably is. Even if it is not,  */
	/* the chances for detecting a song change are better (or at least not  */
	/* worse ;-) than NULL.                                                 */
	if (remove_illegal && need_tag && !strlen(pic_file))
	    return NULL;
	else
	    if (!remove_illegal){
		strcpy(pic_file,xmms_remote_get_playlist_title(0, xmms_remote_get_playlist_pos(0)));
	    }
    }

    return pic_file;
}

char *album_get_front()
{
    char *pic_file=NULL, *mp3_file=NULL;
    static char *old_pic_file=NULL, *old_mp3_file=NULL;

    if(xmms_remote_get_playlist_length(0)==0)
	return NULL;

    mp3_file = xmms_remote_get_playlist_file(0, xmms_remote_get_playlist_pos(0));

    if(mp3_file == NULL)
	return NULL;

    if (old_mp3_file == NULL)
	old_mp3_file = g_strndup("(null)", 1024);

    if (strcmp(mp3_file, old_mp3_file) != 0) {
	g_free(old_mp3_file);
	old_mp3_file = g_strdup(mp3_file);
	pic_file = album_get_pic_file(picname_format_front, TRUE);
	if (pic_file)
	    old_pic_file = g_strdup(pic_file);
    }
    else {
	if (old_pic_file == NULL)
	    return NULL;
	pic_file = g_strdup(old_pic_file);
    }
    return pic_file;
}

char *album_get_back()
{
    char *pic_file=NULL, *mp3_file=NULL;
    static char *old_pic_file=NULL, *old_mp3_file=NULL;

    if(xmms_remote_get_playlist_length(0)==0)
	return NULL;

    mp3_file = xmms_remote_get_playlist_file(0, xmms_remote_get_playlist_pos(0));

    if(mp3_file == NULL)
	return NULL;

    if (old_mp3_file == NULL)
	old_mp3_file = g_strndup("(null)", 1024);

    if (strcmp(mp3_file, old_mp3_file) != 0) {
	g_free(old_mp3_file);
	old_mp3_file = g_strdup(mp3_file);
	pic_file = album_get_pic_file(picname_format_back, TRUE);
	if (pic_file)
	    old_pic_file = g_strdup(pic_file);
    }
    else {
	if (old_pic_file == NULL)
	    return NULL;
	pic_file = g_strdup(old_pic_file);
    }
    return pic_file;
}

char *album_get_title()
{
    char *pic_file=NULL, *mp3_file=NULL;
    static char *old_pic_file=NULL, *old_mp3_file=NULL;

    if(xmms_remote_get_playlist_length(0)==0)
	return NULL;

    mp3_file = xmms_remote_get_playlist_file(0, xmms_remote_get_playlist_pos(0));

    if(mp3_file == NULL)
	return NULL;

    if (old_mp3_file == NULL)
	old_mp3_file = g_strndup("(null)", 1024);

    if (strcmp(mp3_file, old_mp3_file) != 0) {
	g_free(old_mp3_file);
	old_mp3_file = g_strdup(mp3_file);
	pic_file = album_get_pic_file("%t", FALSE);
	if (pic_file)
	    old_pic_file = g_strdup(pic_file);
    }
    else {
	if (old_pic_file == NULL)
	    return NULL;
	pic_file = g_strdup(old_pic_file);
    }
    return pic_file;
}

/* perfom a fancy animation from the start thumbnail to the end thumbnail */
/* currently only a smooth gradient and slide from left are available.    */
/* the speed of the transition is processor dependent.                    */
/* taken from gai-bgswitcher by greg@drakos7.net                          */
void fancy_animation(GdkPixbuf* start, GdkPixbuf* end)
{
    GdkPixbuf *composite;
    int i, a_type;

    if (pref_anim_type==ANIMATION_TYPES-1)
	a_type = (rand() % (ANIMATION_TYPES-1)) + 1; //(ANIMATION_TYPES-1)*((rand()/(RAND_MAX+1.0)))+1;
    else
	a_type = pref_anim_type;

    for (i=0; i<100; i+=5){ 
	composite = gdk_pixbuf_copy(start);

	switch (a_type){
	case 1:		//fade in
	    gdk_pixbuf_composite(end,composite,0, 0, 56, 56, 0, 0,
				 1, 1, ANIMATION_METHOD,(int)(i*255.0/100));
	    break;
	case 2: 	//stretch from left
	    gdk_pixbuf_composite(end,composite,0,0,
				 (int)(56*i/100), 56,0,0,i/(100.0),1.0,ANIMATION_METHOD,255);
	    break;
	case 3:		//stretch from top
	    gdk_pixbuf_composite(end,composite,0,0,
				 56, (int)(56*i/100),0,0,1.0,i/(100.0),ANIMATION_METHOD,255);
	    break;
	case 4: 	//zoom
	    gdk_pixbuf_composite(end,composite, 28-(28*i/100), 28-(28*i/100), 56*(i/100.0), 56*(i/100.0),
				 28-(28*i/100), 28-(28*i/100), i/100.0, i/100.0, ANIMATION_METHOD,255);
	    break;
	}

	gai_draw_bg(composite, 0,0, 56, 56, 4, 4);
	gai_draw_update_bg();
	g_object_unref(composite);
	usleep(4000*(11-ANIMATION_SPEED));
    }
}



void album_update(gpointer d)
{
    GdkPixbuf *tmp_pixbuf=NULL, *bg_image;
    static GdkPixbuf *curr_pixbuf = NULL;
    char *pic=NULL, *pic2=NULL, *song=NULL;
    //    char *tooltip_text;
    static int xmms_working = 1;
    static char *old_pic = NULL, *old_song = NULL;
    static int old_no_image = 0;

    if(old_pic == NULL)
	old_pic = g_strdup("af!\n");
    if(old_song == NULL)
	old_song = g_strdup("af!\n");

    if(!xmms_remote_is_running(0)){
	if(xmms_working || force_update){
		force_update = 0;
		gai_draw_bg(no_xmms, 0, 0, 56, 56, 4, 4);
		gai_draw_update_bg();
		xmms_working = 0;
		no_image = 1;
	}
    } else {
	xmms_working = 1;
	no_image = 0;

	if(front)
	    pic = album_get_front();
	else
	    pic = album_get_back();



	song = album_get_title();

	if(pic == NULL){
	    if (tmp_pixbuf != NULL)
		g_object_unref(tmp_pixbuf);
	    tmp_pixbuf = naif;
	    no_image = 1;
	}
	else if(force_update || (strcmp(song, old_song) && strcmp(pic,old_pic))){

	    force_update = 0;
	    
	    if (!strcmp(song, old_song) && !strcmp(pic, old_pic))
		no_image = old_no_image;
	    else
		old_no_image = no_image;

	    g_free(old_pic);
	    old_pic = pic;

	    g_free(old_song);
	    old_song = song;

	    /*	    tooltip_text = album_get_pic_file("%p [%a] - %t", FALSE);
	    gai_tooltip_set(tooltip_text);
	    g_free(tooltip_text);*/

	    pic = g_strdup_printf("%s/%s",local_dir, old_pic);
	    pic[strlen(pic)-1]='g';
	    pic[strlen(pic)-2]='n';
	    pic[strlen(pic)-3]='p';

	    /* Checking for cached file */
	    if (tmp_pixbuf != NULL)
		g_object_unref(tmp_pixbuf);
	    tmp_pixbuf = no_image ? NULL : gdk_pixbuf_new_from_file(pic, NULL);

	    if(tmp_pixbuf == NULL){
		pic2 = g_strdup_printf("%s/%s",album_dir, old_pic);
		
		if (tmp_pixbuf != NULL)
		    g_object_unref(tmp_pixbuf);
		tmp_pixbuf = no_image ? NULL : gdk_pixbuf_new_from_file(pic2, NULL);
		g_free(pic2);

		if(tmp_pixbuf == NULL){
		    tmp_pixbuf = naif;
		    no_image = 1;
		    old_no_image = 1;
		    force_update = 1;
		    
		} else {
		    bg_image = gdk_pixbuf_scale_simple(tmp_pixbuf, 56, 56, GDK_INTERP_BILINEAR);
		    gdk_pixbuf_save(bg_image,pic,"png",NULL, NULL);
		    g_object_unref(tmp_pixbuf);
		    tmp_pixbuf = bg_image;
		}
	    }
	    if(pref_fancy_animation && curr_pixbuf != NULL && curr_pixbuf != tmp_pixbuf)
		fancy_animation(curr_pixbuf, tmp_pixbuf);
	    gai_draw_bg(tmp_pixbuf, 0,0, 56, 56, 4, 4);
	    gai_draw_update_bg();
	    if(curr_pixbuf != naif && curr_pixbuf != NULL)
		g_object_unref(curr_pixbuf);
	    curr_pixbuf = tmp_pixbuf;
	}
	g_free(pic);
    }
}

void album_button1(int x,int y, gpointer d)
{
    char *pic, *see_pic;
    if(no_image) return;

    pic = album_get_front();
    see_pic = g_strdup_printf("%s \"%s/%s\"",view_prg, album_dir, pic);
    gai_exec(see_pic);
    g_free(pic);
    g_free(see_pic);

}

void album_button2(int x,int y, gpointer d)
{
    char *pic, *see_pic;
    if(no_image) return;

    pic = album_get_back();
    see_pic = g_strdup_printf("%s \"%s/%s\"",view_prg,album_dir, pic);
    gai_exec(see_pic);
    g_free(pic);
    g_free(see_pic);

}

void restart_album(gpointer d)
{
    gai_save_string("album/front", picname_format_front);
    gai_save_string("album/back", picname_format_back);
    gai_save_string("album/dir",album_dir);
    gai_save_string("album/view", view_prg);
    gai_save_int("album/animate", pref_fancy_animation);
    gai_save_int("album/anim_type", pref_anim_type);

}

void album_change(int orient, int width, int height, gpointer d)
{
    force_update = 1;
}

void album_scroll(int dir, gpointer d)
{
    front = !front; 
    force_update = 1;
}

int main(int argc, char **argv)
{

    char *home_dir;

    gai_init2(&applet_defines, &argc, &argv);
    gai_background_set(64, 64, GAI_BACKGROUND_MAX_SIZE_NONE, TRUE);

    home_dir = getenv("HOME");
    local_dir = g_strdup_printf("%s/.gai-album", home_dir);
    mkdir(local_dir,0777);

    picname_format_front = gai_load_string_with_default("album/front", "%p - %a - front.jpg");
    picname_format_back = gai_load_string_with_default("album/back", "%p - %a - back.jpg");
    album_dir = gai_load_string_with_default("album/dir","/usr/local/albums/");
    view_prg = gai_load_string_with_default("album/view", "xv -geometry 640x480");
    pref_fancy_animation= gai_load_int_with_default("album/animate",DEFAULT_ANIMATE);
    pref_anim_type= gai_load_int_with_default("album/anim_type", DEFAULT_ANIM_TYPE);


    no_image = 0;

    no_xmms = gai_load_image("no_xmms.png");
    naif = gai_load_image("naif.png");
    gai_signal_on_update((GaiCallback0 *)album_update,1000,NULL);
    gai_signal_on_mouse_button_click((GaiCallback2 *)album_button1, GAI_MOUSE_BUTTON_1, NULL);
    gai_signal_on_mouse_button_click((GaiCallback2 *)album_button2, GAI_MOUSE_BUTTON_2, NULL);
    gai_signal_on_scroll_buttons((GaiCallback1 *)album_scroll, NULL);
    gai_signal_on_change((GaiCallback3 *)album_change, NULL);

    gai_preferences2("GAI Album - Preferences", album_pref, 
		    help_text,
		    (GaiCallback0 *)restart_album, NULL);


    gai_start();
    g_object_unref(no_xmms);
    g_object_unref(naif);
    return 0;
}
